from collections import defaultdict
from dataclasses import dataclass
from typing import Generator

from infrahub.core.constants import DiffAction
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.query.node import NodeGetKindQuery
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger

from ..model.path import (
    CalculatedDiffs,
    EnrichedDiffConflict,
    EnrichedDiffProperty,
    EnrichedDiffRelationship,
    EnrichedDiffRoot,
)
from ..payload_builder import get_display_labels
from .interface import DiffEnricherInterface

log = get_logger()

PROPERTY_TYPES_WITH_LABELS = {DatabaseEdgeType.IS_RELATED, DatabaseEdgeType.HAS_OWNER, DatabaseEdgeType.HAS_SOURCE}


@dataclass
class DisplayLabelRequest:
    node_id: str
    branch_name: str

    def __hash__(self) -> int:
        return hash(f"{self.node_id}-{self.branch_name}")


class DiffLabelsEnricher(DiffEnricherInterface):
    """Add display labels for nodes and labels for relationships"""

    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db
        self._base_branch_name: str | None = None
        self._diff_branch_name: str | None = None
        self._conflicts_only = False

    @property
    def base_branch_name(self) -> str:
        if not self._base_branch_name:
            raise RuntimeError("could not identify base branch")
        return self._base_branch_name

    @property
    def diff_branch_name(self) -> str:
        if not self._diff_branch_name:
            raise RuntimeError("could not identify diff branch")
        return self._diff_branch_name

    def _get_branch_for_action(self, action: DiffAction) -> str:
        if action is DiffAction.REMOVED:
            return self.base_branch_name
        return self.diff_branch_name

    def _nodes_iterator(self, enriched_diff_root: EnrichedDiffRoot) -> Generator[DisplayLabelRequest, str | None, None]:
        for node in enriched_diff_root.nodes:
            if not self._conflicts_only:
                branch_name = self._get_branch_for_action(action=node.action)
                label = yield DisplayLabelRequest(node_id=node.uuid, branch_name=branch_name)
                if label:
                    node.label = label
            for attribute_diff in node.attributes:
                for property_diff in attribute_diff.properties:
                    property_iterator = self._property_iterator(property_diff=property_diff)
                    try:
                        label = None
                        while True:
                            display_label_request = property_iterator.send(label)
                            label = yield display_label_request
                    except StopIteration:
                        ...
            for relationship_diff in node.relationships:
                relationship_iterator = self._relationship_iterator(relationship_diff=relationship_diff)
                try:
                    label = None
                    while True:
                        display_label_request = relationship_iterator.send(label)
                        label = yield display_label_request
                except StopIteration:
                    ...

    def _relationship_iterator(
        self, relationship_diff: EnrichedDiffRelationship
    ) -> Generator[DisplayLabelRequest, str | None, None]:
        for element_diff in relationship_diff.relationships:
            if not self._conflicts_only:
                branch_name = self._get_branch_for_action(action=element_diff.action)
                peer_label = yield DisplayLabelRequest(node_id=element_diff.peer_id, branch_name=branch_name)
                if peer_label:
                    element_diff.peer_label = peer_label
            if element_diff.conflict:
                conflict_iterator = self._conflict_iterator(conflict_diff=element_diff.conflict)
                label = None
                try:
                    while True:
                        display_label_request = conflict_iterator.send(label)
                        label = yield display_label_request
                except StopIteration:
                    ...
            for property_diff in element_diff.properties:
                property_iterator = self._property_iterator(property_diff=property_diff)
                label = None
                try:
                    while True:
                        display_label_request = property_iterator.send(label)
                        label = yield display_label_request
                except StopIteration:
                    ...

    def _property_iterator(
        self, property_diff: EnrichedDiffProperty
    ) -> Generator[DisplayLabelRequest, str | None, None]:
        if property_diff.property_type in PROPERTY_TYPES_WITH_LABELS:
            if property_diff.previous_value and not self._conflicts_only:
                label = yield DisplayLabelRequest(
                    node_id=property_diff.previous_value, branch_name=self.base_branch_name
                )
                if label:
                    property_diff.previous_label = label
            if property_diff.new_value and not self._conflicts_only:
                label = yield DisplayLabelRequest(node_id=property_diff.new_value, branch_name=self.diff_branch_name)
                if label:
                    property_diff.new_label = label
            if property_diff.conflict:
                conflict_iterator = self._conflict_iterator(conflict_diff=property_diff.conflict)
                label = None
                try:
                    while True:
                        display_label_request = conflict_iterator.send(label)
                        label = yield display_label_request
                except StopIteration:
                    ...

    def _conflict_iterator(
        self, conflict_diff: EnrichedDiffConflict
    ) -> Generator[DisplayLabelRequest, str | None, None]:
        if conflict_diff.base_branch_value:
            label = yield DisplayLabelRequest(
                node_id=conflict_diff.base_branch_value, branch_name=self.base_branch_name
            )
            if label:
                conflict_diff.base_branch_label = label
        if conflict_diff.diff_branch_value:
            label = yield DisplayLabelRequest(
                node_id=conflict_diff.diff_branch_value, branch_name=self.diff_branch_name
            )
            if label:
                conflict_diff.diff_branch_label = label

    def _update_relationship_labels(self, enriched_diff: EnrichedDiffRoot) -> None:
        for node in enriched_diff.nodes:
            if not node.relationships:
                continue

            node_schema = self.db.schema.get(name=node.kind, branch=enriched_diff.diff_branch_name, duplicate=False)
            alternate_node_schema = None
            if enriched_diff.diff_branch_name != enriched_diff.base_branch_name and self.db.schema.has(
                name=node.kind, branch=enriched_diff.base_branch_name
            ):
                alternate_node_schema = self.db.schema.get(
                    name=node.kind, branch=enriched_diff.base_branch_name, duplicate=False
                )
            for relationship_diff in node.relationships:
                relationship_schema = node_schema.get_relationship_or_none(name=relationship_diff.name)
                if not relationship_schema and alternate_node_schema:
                    relationship_schema = alternate_node_schema.get_relationship_or_none(name=relationship_diff.name)
                relationship_diff.label = relationship_schema.label or "" if relationship_schema else ""

    async def _get_display_label_map(
        self, display_label_requests: set[DisplayLabelRequest]
    ) -> dict[str, dict[str, str]]:
        node_ids = [dlr.node_id for dlr in display_label_requests]
        query = await NodeGetKindQuery.init(db=self.db, ids=node_ids)
        await query.execute(db=self.db)
        node_kind_map = await query.get_node_kind_map()
        display_label_request_map: dict[str, dict[str, list[str]]] = defaultdict(dict)
        for dlr in display_label_requests:
            try:
                node_kind = node_kind_map[dlr.node_id]
            except KeyError:
                continue
            branch_map = display_label_request_map[dlr.branch_name]
            if node_kind not in branch_map:
                branch_map[node_kind] = []
            branch_map[node_kind].append(dlr.node_id)
        return await get_display_labels(db=self.db, nodes=display_label_request_map)

    async def enrich(
        self,
        enriched_diff_root: EnrichedDiffRoot,
        calculated_diffs: CalculatedDiffs | None = None,  # noqa: ARG002
        conflicts_only: bool = False,
    ) -> None:
        log.info("Beginning display labels diff enrichment...")
        self._base_branch_name = enriched_diff_root.base_branch_name
        self._diff_branch_name = enriched_diff_root.diff_branch_name
        self._conflicts_only = conflicts_only
        display_label_requests = set(self._nodes_iterator(enriched_diff_root=enriched_diff_root))
        display_label_map = await self._get_display_label_map(display_label_requests=display_label_requests)

        # iterate through all the labels in this diff again and set them, if possible
        nodes_iterator = self._nodes_iterator(enriched_diff_root=enriched_diff_root)
        try:
            display_label_request = next(nodes_iterator)
            while display_label_request:
                try:
                    display_label = display_label_map[display_label_request.branch_name][display_label_request.node_id]
                except KeyError:
                    display_label = None
                display_label_request = nodes_iterator.send(display_label)
        except StopIteration:
            ...

        self._update_relationship_labels(enriched_diff=enriched_diff_root)
        log.info("Display labels diff enrichment complete.")



from typing import Any

from infrahub.core.attribute import BaseAttribute
from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.node import Node
from infrahub.core.relationship import RelationshipManager
from infrahub.core.relationship.model import Relationship
from infrahub.database import InfrahubDatabase

from .queries.get_profile_data import GetProfileDataQuery, ProfileData, RelationshipFilter


class NodeProfilesApplier:
    """Applies profile values to nodes and templates.

    Profile values take precedence over both default values and template-sourced values.
    When a template has profiles assigned:
    1. Profile values are applied to the template itself
    2. Nodes created from that template inherit the profile values (not the template's own values)
    3. Profile priority determines which profile wins when multiple profiles set the same attribute
    """

    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch

    async def _get_profile_ids(self, node: Node) -> list[str]:
        try:
            profiles_rel = node.get_relationship("profiles")
        except ValueError:
            return []
        profile_rels = await profiles_rel.get_relationships(db=self.db)
        return [pr.peer_id for pr in profile_rels if pr.peer_id]

    async def _get_attr_names_for_profiles(self, node: Node) -> list[str]:
        node_schema = node.get_schema()

        # get the names of attributes that could be affected by profile changes
        attr_names_for_profiles: list[str] = []
        for attr_schema in node_schema.attributes:
            attr_name = attr_schema.name
            node_attr: BaseAttribute = getattr(node, attr_name)
            if node_attr.is_from_profile or node_attr.is_default:
                attr_names_for_profiles.append(attr_name)
        return attr_names_for_profiles

    async def _get_rel_names_for_profiles(self, node: Node) -> list[str]:
        node_schema = node.get_schema()

        rel_names_for_profiles: list[str] = []
        for rel_schema in node_schema.relationships:
            if not rel_schema.support_profiles:
                continue

            rel_name = rel_schema.name
            node_rel = node.get_relationship(rel_name)

            current_rels = await node_rel.get_relationships(db=self.db)
            if node_rel.is_from_profile or len(current_rels) == 0:
                rel_names_for_profiles.append(rel_name)

        return rel_names_for_profiles

    async def _get_rel_filters_for_profiles(self, node: Node, rel_names: list[str]) -> list[RelationshipFilter]:
        node_schema = node.get_schema()

        identifiers: list[RelationshipFilter] = []
        for rel_name in rel_names:
            rel_schema = node_schema.get_relationship(name=rel_name)
            identifiers.append(
                RelationshipFilter(
                    relationship_identifier=f"profile_{rel_schema.get_identifier()}", direction=rel_schema.direction
                )
            )

        return identifiers

    async def _get_sorted_profile_data(
        self,
        profile_ids: list[str],
        attr_names_for_profiles: list[str],
        relationship_filters: list[RelationshipFilter] | None = None,
    ) -> list[ProfileData]:
        if not profile_ids:
            return []
        query = await GetProfileDataQuery.init(
            db=self.db,
            branch=self.branch,
            profile_ids=profile_ids,
            attr_names=attr_names_for_profiles,
            relationship_filters=relationship_filters,
        )
        await query.execute(db=self.db)
        profile_data_list = query.get_profile_data()
        return sorted(profile_data_list, key=lambda x: (x.priority, x.uuid))

    def _apply_profile_to_attribute(self, node_attr: BaseAttribute, profile_value: Any, profile_id: str) -> bool:
        is_changed = False
        if node_attr.value != profile_value:
            node_attr.value = profile_value
            is_changed = True
        if node_attr.is_default is not False:
            node_attr.is_default = False
            is_changed = True
        if node_attr.is_from_profile is not True:
            node_attr.is_from_profile = True
            is_changed = True
        if node_attr.source_id != profile_id:  # type: ignore[attr-defined]
            node_attr.set_source(value=profile_id)
            is_changed = True
        return is_changed

    def _remove_profile_from_attribute(self, node_attr: BaseAttribute) -> None:
        node_attr.clear_source()
        node_attr.value = node_attr.schema.default_value
        node_attr.is_default = True
        node_attr.is_from_profile = False

    async def _apply_profile_to_relationship(
        self, node: Node, node_rel: RelationshipManager, peer_ids: list[str], profile_id: str
    ) -> bool:
        """Apply profile relationship peers to a node relationship.

        Profile relationships are only applied if the node has no user-set peers for this relationship.
        User-set peers (not from profile) take precedence over profile values.
        """
        is_changed = False

        current_rels = await node_rel.get_relationships(db=self.db)
        profile_peer_ids = set(peer_ids)

        user_set_peer_ids: set[str] = set()
        profile_set_rels: list[Relationship] = []
        for rel in current_rels:
            if rel.peer_id:
                if rel.is_from_profile:
                    profile_set_rels.append(rel)
                else:
                    user_set_peer_ids.add(rel.peer_id)

        # If user has set any peers, they override profile values entirely
        if user_set_peer_ids:
            for rel in profile_set_rels:
                if rel.peer_id and rel.peer_id not in user_set_peer_ids:
                    await node_rel.remove_locally(peer_id=rel.peer_id, db=self.db)
                    is_changed = True
                elif rel.peer_id and rel.peer_id in user_set_peer_ids:
                    rel.clear_source()
                    is_changed = True

            if is_changed:
                node_rel.is_from_profile = False
                await node_rel.save(db=self.db)
            return is_changed

        profile_set_peer_ids = {rel.peer_id for rel in profile_set_rels if rel.peer_id}
        relationships_to_replace: dict[str, list[Relationship]] = {"insert": [], "remove": []}

        # Remove relationships that are from profile but not in the new profile value
        for rel in profile_set_rels:
            if rel.peer_id and rel.peer_id not in profile_peer_ids:
                relationships_to_replace["remove"].append(rel)

        # Add relationships that are in profile but not present
        for peer_id in profile_peer_ids:
            if peer_id not in profile_set_peer_ids:
                new_rel = Relationship(
                    schema=node_rel.schema,
                    branch=self.branch,
                    source_kind=InfrahubKind.PROFILE,
                    node=node,
                )
                await new_rel.new(db=self.db, data=peer_id)
                new_rel.set_source(value=profile_id)
                relationships_to_replace["insert"].append(new_rel)

        is_changed |= await node_rel.update_relationships(
            db=self.db,
            relationships_to_remove=relationships_to_replace["remove"],
            relationships_to_insert=relationships_to_replace["insert"],
        )

        if profile_peer_ids:
            node_rel.is_from_profile = True

        if is_changed:
            await node_rel.save(db=self.db)

        return is_changed

    async def _remove_profile_from_relationship(self, relationship_manager: RelationshipManager) -> None:
        relationship_manager.is_from_profile = False
        await relationship_manager.delete(db=self.db)

    async def apply_profiles(self, node: Node) -> list[str]:
        profile_ids = await self._get_profile_ids(node=node)
        attr_names_for_profiles = await self._get_attr_names_for_profiles(node=node)
        rel_names_for_profiles = await self._get_rel_names_for_profiles(node=node)
        rel_filters_for_profiles = await self._get_rel_filters_for_profiles(node=node, rel_names=rel_names_for_profiles)
        if not attr_names_for_profiles and not rel_filters_for_profiles:
            return []

        # get profiles priorities, attribute values, and relationship peers on branch
        sorted_profile_data = await self._get_sorted_profile_data(
            profile_ids=profile_ids,
            attr_names_for_profiles=attr_names_for_profiles,
            relationship_filters=rel_filters_for_profiles,
        )

        updated_field_names: list[str] = []
        # set attribute values/is_default/is_from_profile on nodes
        for attr_name in attr_names_for_profiles:
            has_profile_attr_data = False
            node_attr = node.get_attribute(attr_name)
            for profile_data in sorted_profile_data:
                profile_value = profile_data.attribute_values.get(attr_name)
                if profile_value is not None:
                    has_profile_attr_data = True
                    is_changed = self._apply_profile_to_attribute(
                        node_attr=node_attr, profile_value=profile_value, profile_id=profile_data.uuid
                    )
                    if is_changed:
                        updated_field_names.append(attr_name)
                    break
            if not has_profile_attr_data and node_attr.is_from_profile:
                self._remove_profile_from_attribute(node_attr=node_attr)
                updated_field_names.append(attr_name)

        for rel_filter in rel_filters_for_profiles:
            has_profile_rel_data = False
            node_rel = node.get_relationship_by_identifier(rel_filter.relationship_identifier.removeprefix("profile_"))

            for profile_data in sorted_profile_data:
                profile_peers = profile_data.relationship_peers.get(rel_filter)
                if profile_peers:
                    has_profile_rel_data = True
                    is_changed = await self._apply_profile_to_relationship(
                        node=node, node_rel=node_rel, peer_ids=profile_peers, profile_id=profile_data.uuid
                    )
                    if is_changed:
                        updated_field_names.append(node_rel.name)
                    break

            # Refresh the relationship manager to update the is_from_profile property
            await node_rel.fetch_relationship_ids(db=self.db)
            if not has_profile_rel_data and node_rel.is_from_profile:
                await self._remove_profile_from_relationship(relationship_manager=node_rel)
                updated_field_names.append(node_rel.name)

        return updated_field_names



from __future__ import annotations

from .constants import HFID_STR_SEPARATOR


def parse_human_friendly_id(hfid: str | list[str]) -> tuple[str | None, list[str]]:
    """Parse a human-friendly ID into a kind and an identifier."""
    if isinstance(hfid, str):
        hfid_parts = hfid.split(HFID_STR_SEPARATOR)
        if len(hfid_parts) == 1:
            return None, hfid_parts
        return hfid_parts[0], hfid_parts[1:]
    if isinstance(hfid, list):
        return None, hfid
    raise ValueError(f"Invalid human friendly ID: {hfid}")



from enum import Enum, Flag, auto

PARENT_CHILD_IDENTIFIER = "parent__child"
RESOURCE_POOL_REL_SUFFIX = "_from_resource_pool"


class FlagProperty(Enum):
    IS_PROTECTED = "is_protected"


class NodeProperty(Enum):
    SOURCE = "source"
    OWNER = "owner"


class UpdateSupport(Enum):
    NOT_SUPPORTED = "not_supported"
    ALLOWED = "allowed"
    MIGRATION_REQUIRED = "migration_required"
    VALIDATE_CONSTRAINT = "validate_constraint"
    NOT_APPLICABLE = "not_applicable"


class UpdateValidationErrorType(Enum):
    NOT_SUPPORTED = "not_supported"
    VALIDATOR_FAILED = "validator_failed"
    MIGRATION_NOT_AVAILABLE = "migration_not_available"
    VALIDATOR_NOT_AVAILABLE = "validator_not_available"


class SchemaElementPathType(Flag):
    ATTR_WITH_PROP = auto()
    ATTR_NO_PROP = auto()
    ATTR = ATTR_WITH_PROP | ATTR_NO_PROP

    REL_MANY_NO_ATTR = auto()
    REL_MANY_ATTR = auto()
    REL_MANY = REL_MANY_NO_ATTR | REL_MANY_ATTR

    REL_ONE_MANDATORY_NO_ATTR = auto()
    REL_ONE_MANDATORY_ATTR_WITH_PROP = auto()
    REL_ONE_MANDATORY_ATTR_NO_PROP = auto()
    REL_ONE_MANDATORY_ATTR = REL_ONE_MANDATORY_ATTR_WITH_PROP | REL_ONE_MANDATORY_ATTR_NO_PROP
    REL_ONE_MANDATORY = REL_ONE_MANDATORY_ATTR | REL_ONE_MANDATORY_NO_ATTR
    REL_ONE_OPTIONAL_ATTR_WITH_PROP = auto()
    REL_ONE_OPTIONAL_ATTR_NO_PROP = auto()
    REL_ONE_OPTIONAL_ATTR = REL_ONE_OPTIONAL_ATTR_WITH_PROP | REL_ONE_OPTIONAL_ATTR_NO_PROP
    REL_ONE_OPTIONAL_NO_ATTR = auto()
    REL_ONE_OPTIONAL = REL_ONE_OPTIONAL_ATTR | REL_ONE_OPTIONAL_NO_ATTR
    REL_ONE = REL_ONE_MANDATORY | REL_ONE_OPTIONAL
    REL_ONE_ATTR_WITH_PROP = REL_ONE_MANDATORY_ATTR_WITH_PROP | REL_ONE_OPTIONAL_ATTR_WITH_PROP
    REL_ONE_ATTR = REL_ONE_MANDATORY_ATTR | REL_ONE_OPTIONAL_ATTR
    REL_ONE_NO_ATTR = REL_ONE_MANDATORY_NO_ATTR | REL_ONE_OPTIONAL_NO_ATTR

    REL = REL_ONE | REL_MANY
    REL_ATTR = REL_ONE_MANDATORY_ATTR | REL_ONE_OPTIONAL_ATTR | REL_MANY_ATTR
    RELS_NO_ATTR = REL_ONE_MANDATORY_NO_ATTR | REL_ONE_OPTIONAL_NO_ATTR | REL_MANY_NO_ATTR



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..shared import GraphMigration, MigrationInput

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class BackfillAttributeMetadataQuery(Query):
    """Backfill created_at and updated_at for Attribute vertices on default/global branches.

    - created_at = from time of HAS_ATTRIBUTE edge
    - updated_at = latest from or to time of any non-HAS_ATTRIBUTE edge
    """

    name = "backfill_attribute_metadata"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        // Find all Attribute vertices on default/global branch without created_at
        MATCH (:Node)-[ha:HAS_ATTRIBUTE {status: "active"}]->(attr:Attribute)
        WHERE ha.branch_level = 1
        AND attr.created_at IS NULL

        // created_at = from time of HAS_ATTRIBUTE edge
        WITH DISTINCT attr, ha.from AS created_at

        // Find latest change time from all non-HAS_ATTRIBUTE edges (both from and to)
        CALL (attr) {
            MATCH (attr)-[e]->()
            WHERE e.branch_level = 1
            WITH e.from AS change_time
            WHERE change_time IS NOT NULL
            RETURN change_time
            UNION ALL
            MATCH (attr)-[e]->()
            WHERE e.branch_level = 1
            AND e.to IS NOT NULL
            RETURN e.to AS change_time
        }
        WITH attr, created_at, max(change_time) AS updated_at

        // Set metadata (use coalesce to fall back to created_at if no changes found)
        CALL (attr, created_at, updated_at) {
            SET attr.created_at = created_at
            SET attr.updated_at = coalesce(updated_at, created_at)
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class BackfillRelationshipMetadataQuery(Query):
    """Backfill created_at and updated_at for Relationship vertices on default/global branches.

    - created_at = earliest from time of any IS_RELATED edge
    - updated_at = latest from or to time of any non-IS_RELATED edge
    """

    name = "backfill_relationship_metadata"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        // Find all Relationship vertices on default/global branch without created_at
        MATCH (:Node)-[ir:IS_RELATED]-(rel:Relationship)
        WHERE ir.branch_level = 1
        AND rel.created_at IS NULL
        WITH DISTINCT rel

        // created_at = earliest IS_RELATED edge from time
        CALL (rel) {
            MATCH ()-[ir:IS_RELATED]-(rel)
            WHERE ir.branch_level = 1
            RETURN min(ir.from) AS created_at
        }

        // Find latest change time from all non-IS_RELATED edges (both from and to)
        CALL (rel) {
            MATCH (rel)-[e:!IS_RELATED]-()
            WHERE e.branch_level = 1
            WITH e.from AS change_time
            WHERE change_time IS NOT NULL
            RETURN change_time
            UNION ALL
            MATCH (rel)-[e:!IS_RELATED]-()
            WHERE e.branch_level = 1
            AND e.to IS NOT NULL
            RETURN e.to AS change_time
        }
        WITH rel, created_at, max(change_time) AS updated_at

        // Set metadata (use coalesce to fall back to created_at if no changes found)
        CALL (rel, created_at, updated_at) {
            SET rel.created_at = created_at
            SET rel.updated_at = coalesce(updated_at, created_at)
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class BackfillNodeMetadataQuery(Query):
    """Backfill created_at and updated_at for Node vertices on default/global branches.

    - created_at = earliest from time of any IS_PART_OF edge for Nodes with this UUID
    - updated_at = latest updated_at from any linked Attribute or Relationship

    Must run after BackfillAttributeMetadataQuery and BackfillRelationshipMetadataQuery
    since Node.updated_at depends on field updated_at values.
    """

    name = "backfill_node_metadata"
    type: QueryType = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        // Find all Node vertices on default/global branch without created_at
        MATCH (n:Node)-[e:IS_PART_OF]->(:Root)
        WHERE e.branch_level = 1
        AND n.created_at IS NULL
        WITH DISTINCT n

        // created_at = earliest IS_PART_OF edge from time for this UUID
        // (handles migrated kind/namespace where multiple Node vertices share same UUID)
        CALL (n) {
            MATCH (:Node {uuid: n.uuid})-[ip:IS_PART_OF]->(:Root)
            WHERE ip.branch_level = 1
            RETURN min(ip.from) AS created_at
        }

        // updated_at = latest updated_at from any linked Attribute or Relationship
        CALL (n) {
            MATCH (n)-[:HAS_ATTRIBUTE|IS_RELATED]-(field:Attribute|Relationship)
            WHERE field.updated_at IS NOT NULL
            RETURN max(field.updated_at) AS latest_field_update
        }

        // Set metadata (use coalesce to fall back to created_at if no field updates found)
        CALL (n, created_at, latest_field_update) {
            SET n.created_at = created_at
            SET n.updated_at = coalesce(latest_field_update, created_at)
        } IN TRANSACTIONS
        """
        self.add_to_query(query)


class Migration050(GraphMigration):
    name: str = "050_backfill_vertex_metadata"
    minimum_version: int = 49
    queries: Sequence[type[Query]] = [
        BackfillAttributeMetadataQuery,  # Run first
        BackfillRelationshipMetadataQuery,  # Run second
        BackfillNodeMetadataQuery,  # Run last (depends on Attr/Rel updated_at)
    ]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:
        return await self.do_execute(migration_input=migration_input)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from infrahub.constants.database import IndexType
from infrahub.core.attribute import MAX_STRING_LENGTH
from infrahub.core.migrations.shared import MigrationInput, MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType
from infrahub.database.index import IndexItem
from infrahub.database.neo4j import IndexManagerNeo4j
from infrahub.log import get_logger

from ..shared import ArbitraryMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


AV_INDEXED_INDEX = IndexItem(
    name="attr_value_indexed", label="AttributeValueIndexed", properties=["value"], type=IndexType.RANGE
)


@dataclass
class SchemaAttributeTimeframe:
    kind: str
    attr_name: str
    branch: str
    branch_level: int
    branched_from: str
    from_time: str
    is_default_branch: bool
    is_large_type: bool


class GetLargeAttributeTypesQuery(Query):
    """For every active attribute on every branch, return a SchemaAttributeTimeframe object"""

    name = "get_large_attribute_types_query"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (branch:Branch)
// --------------
// find all SchemaAttributes with a LARGE_ATTRIBUTE_TYPE kind for each branch
// --------------
MATCH (schema_attr:SchemaAttribute)-[r1:HAS_ATTRIBUTE]->(attr_kind:Attribute {name: "kind"})-[r2:HAS_VALUE]->(attr_kind_value)
WHERE r1.status = "active" and r1.to IS NULL AND r2.status = "active" and r2.to IS NULL
WITH DISTINCT branch, schema_attr, attr_kind
CALL (schema_attr, attr_kind, branch) {
    MATCH (schema_attr)-[has_attr:HAS_ATTRIBUTE]->(attr_kind)-[has_value:HAS_VALUE]->(attr_kind_value)
    WHERE has_attr.status = "active"
    AND has_value.status = "active"
    AND has_attr.to IS NULL
    AND has_value.to IS NULL
    AND (
        has_attr.branch = branch.name
        OR (has_attr.branch_level < branch.hierarchy_level AND has_attr.from <= branch.branched_from)
    )
    AND (
        has_value.branch = branch.name
        OR (has_value.branch_level < branch.hierarchy_level AND has_value.from <= branch.branched_from)
    )
    WITH has_value.from AS from_time, attr_kind_value.value AS attr_type
    ORDER BY has_value.branch_level DESC, has_value.from DESC
    LIMIT 1
    WITH from_time, attr_type
    RETURN from_time, attr_type IN ["JSON", "List", "TextArea"] AS is_large_type
}
CALL (schema_attr, branch) {
    // --------------
    // get the attribute name
    // --------------
    MATCH (schema_attr)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(name_value)
    WHERE r1.status = "active"
    AND r1.to IS NULL
    AND (
        r1.branch = branch.name
        OR (r1.branch_level < branch.hierarchy_level AND r1.from <= branch.branched_from)
    )
    AND r2.status = "active"
    AND r2.to IS NULL
    AND (
        r2.branch = branch.name
        OR (r2.branch_level < branch.hierarchy_level AND r2.from <= branch.branched_from)
    )
    WITH name_value.value AS attr_name
    ORDER BY r2.branch_level DESC, r1.branch_level DESC, r2.from DESC, r1.from DESC
    LIMIT 1

    // --------------
    // get the the schema node/generic
    // --------------
    MATCH (schema_attr)-[r1:IS_RELATED]-(:Relationship {name: "schema__node__attributes"})-[r2:IS_RELATED]-(schema_node:SchemaNode|SchemaGeneric)
    WHERE r1.status = "active"
    AND r1.to IS NULL
    AND (
        r1.branch = branch.name
        OR (r1.branch_level < branch.hierarchy_level AND r1.from <= branch.branched_from)
    )
    AND r2.status = "active"
    AND r2.to IS NULL
    AND (
        r2.branch = branch.name
        OR (r2.branch_level < branch.hierarchy_level AND r2.from <= branch.branched_from)
    )
    WITH attr_name, schema_node
    ORDER BY r2.branch_level DESC, r1.branch_level DESC, r2.from DESC, r1.from DESC
    LIMIT 1

    // --------------
    // find the namespace for this SchemaNode/SchemaGeneric
    // --------------
    MATCH (schema_node)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[r2:HAS_VALUE]->(kind_namespace_value)
    WHERE r1.status = "active"
    AND r1.to IS NULL
    AND (
        r1.branch = branch.name
        OR (r1.branch_level < branch.hierarchy_level AND r1.from <= branch.branched_from)
    )
    AND r2.status = "active"
    AND r2.to IS NULL
    AND (
        r2.branch = branch.name
        OR (r2.branch_level < branch.hierarchy_level AND r2.from <= branch.branched_from)
    )
    WITH attr_name, schema_node, kind_namespace_value.value AS kind_namespace
    ORDER BY r2.branch_level DESC, r1.branch_level DESC, r2.from DESC, r1.from DESC
    LIMIT 1

    // --------------
    // find the name for this SchemaNode/SchemaGeneric
    // --------------
    MATCH (schema_node)-[r1:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[r2:HAS_VALUE]->(kind_name_value)
    WHERE r1.status = "active"
    AND r1.to IS NULL
    AND (
        r1.branch = branch.name
        OR (r1.branch_level < branch.hierarchy_level AND r1.from <= branch.branched_from)
    )
    AND r2.status = "active"
    AND r2.to IS NULL
    AND (
        r2.branch = branch.name
        OR (r2.branch_level < branch.hierarchy_level AND r2.from <= branch.branched_from)
    )
    WITH attr_name, kind_namespace, kind_name_value.value AS kind_name
    ORDER BY r2.branch_level DESC, r1.branch_level DESC, r2.from DESC, r1.from DESC
    LIMIT 1
    RETURN attr_name, kind_namespace, kind_name
}
RETURN
    kind_namespace,
    kind_name,
    attr_name,
    branch.name AS branch,
    branch.hierarchy_level AS branch_level,
    branch.branched_from AS branched_from,
    branch.is_default AS is_default_branch,
    from_time,
    is_large_type
"""
        self.add_to_query(query)
        self.return_labels = [
            "kind_namespace",
            "kind_name",
            "attr_name",
            "branch",
            "branch_level",
            "branched_from",
            "is_default_branch",
            "from_time",
            "is_large_type",
        ]

    def get_large_attribute_type_timeframes(self) -> list[SchemaAttributeTimeframe]:
        schema_attribute_timeframes: list[SchemaAttributeTimeframe] = []
        for result in self.get_results():
            kind_namespace = result.get_as_type("kind_namespace", return_type=str)
            kind_name = result.get_as_type("kind_name", return_type=str)
            attr_name = result.get_as_type("attr_name", return_type=str)
            branch = result.get_as_type("branch", return_type=str)
            branch_level = result.get_as_type("branch_level", return_type=int)
            branched_from = result.get_as_type("branched_from", return_type=str)
            is_default_branch = result.get_as_type("is_default_branch", return_type=bool)
            is_large_type = result.get_as_type("is_large_type", return_type=bool)
            from_time = result.get_as_type("from_time", return_type=str)
            kind = f"{kind_namespace}{kind_name}"
            schema_attribute_timeframes.append(
                SchemaAttributeTimeframe(
                    kind=kind,
                    attr_name=attr_name,
                    branch=branch,
                    branch_level=branch_level,
                    branched_from=branched_from,
                    from_time=from_time,
                    is_default_branch=is_default_branch,
                    is_large_type=is_large_type,
                )
            )
        return schema_attribute_timeframes


class DeIndexLargeAttributeValuesQuery(Query):
    name = "de_index_large_attribute_values_query"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, max_value_size: int, **kwargs: Any) -> None:
        self.max_value_size = max_value_size
        super().__init__(**kwargs)

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["max_value_size"] = self.max_value_size
        query = """
MATCH (av:AttributeValueIndexed)
WHERE size(toString(av.value)) > $max_value_size
REMOVE av:AttributeValueIndexed
        """
        self.add_to_query(query)


class CreateNonIndexedAttributeValueQuery(Query):
    name = "create_non_indexed_attribute_value_query"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        schema_attribute_timeframe: SchemaAttributeTimeframe,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.schema_attribute_timeframe = schema_attribute_timeframe

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params.update(
            {
                "attribute_name": self.schema_attribute_timeframe.attr_name,
                "branch": self.schema_attribute_timeframe.branch,
                "branch_level": self.schema_attribute_timeframe.branch_level,
                "branched_from": self.schema_attribute_timeframe.branched_from,
                "from_time": self.schema_attribute_timeframe.from_time,
            }
        )
        query = """
MATCH (node:Node:%(schema_kind)s)
CALL (node) {
    MATCH (node)-[has_attr_e:HAS_ATTRIBUTE]->(attr:Attribute)
    WHERE attr.name = $attribute_name
    AND (
        has_attr_e.branch = $branch
        OR (has_attr_e.branch_level < $branch_level AND has_attr_e.from <= $branched_from)
    )
    AND has_attr_e.status = "active"
    AND has_attr_e.to IS NULL
    WITH attr
    ORDER BY has_attr_e.branch_level DESC, has_attr_e.from DESC
    LIMIT 1

    // --------------
    // identify the active HAS_VALUE edges that we need to consider
    // --------------
    MATCH (attr)-[has_val_e:HAS_VALUE]->(av)
    WHERE (
        has_val_e.branch = $branch
        OR (has_val_e.branch_level < $branch_level AND has_val_e.from <= $branched_from)
    )
    AND has_val_e.status = "active"
    AND has_val_e.to IS NULL
    RETURN attr, has_val_e, av
    ORDER BY has_val_e.branch_level DESC, has_val_e.from DESC
    LIMIT 1
}
// --------------------
// determine the timestamp to de-index the AttributeValue
// --------------------
WITH attr, has_val_e, av,
CASE
    WHEN $from_time <= has_val_e.from THEN has_val_e.from
    ELSE $from_time
END AS non_indexed_from

// --------------------
// create the new edge to the AttributeValueNonIndexed vertex, if necessary
// --------------------
WITH attr, has_val_e, av, non_indexed_from
CALL (attr, has_val_e, av, non_indexed_from) {
    WITH has_val_e
    WHERE NOT "AttributeValueNonIndexed" IN labels(av)

    MERGE (av_no_index:AttributeValueNonIndexed {value: av.value, is_default: av.is_default})
    LIMIT 1

    CREATE (attr)-[add_no_index_on_branch:HAS_VALUE]->(av_no_index)
    SET add_no_index_on_branch = properties(has_val_e)
    SET
        add_no_index_on_branch.branch = $branch,
        add_no_index_on_branch.branch_level = $branch_level,
        add_no_index_on_branch.from = non_indexed_from,
        add_no_index_on_branch.status = "active",
        add_no_index_on_branch.to = NULL

    // --------------------
    // delete existing active edge if it is on this branch and we created a new edge
    // --------------------
    WITH has_val_e
    WHERE has_val_e.branch = $branch
    DELETE has_val_e

}
        """ % {"schema_kind": self.schema_attribute_timeframe.kind}
        self.add_to_query(query)


class RevertNonIndexOnBranchQuery(Query):
    name = "revert_non_index_on_branch_query"
    type = QueryType.WRITE
    insert_return = False

    def __init__(
        self,
        schema_attribute_timeframe: SchemaAttributeTimeframe,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.schema_attribute_timeframe = schema_attribute_timeframe

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params.update(
            {
                "attribute_name": self.schema_attribute_timeframe.attr_name,
                "branch": self.schema_attribute_timeframe.branch,
                "branch_level": self.schema_attribute_timeframe.branch_level,
                "branched_from": self.schema_attribute_timeframe.branched_from,
                "from_time": self.schema_attribute_timeframe.from_time,
            }
        )
        query = """
MATCH (node:Node:%(schema_kind)s)
CALL (node) {
    MATCH (node)-[has_attr_e:HAS_ATTRIBUTE]->(attr:Attribute)
    WHERE attr.name = $attribute_name
    AND (
        has_attr_e.branch = $branch
        OR (has_attr_e.branch_level < $branch_level AND has_attr_e.from <= $branched_from)
    )
    AND has_attr_e.status = "active"
    AND has_attr_e.to IS NULL
    WITH attr
    ORDER BY has_attr_e.branch_level DESC, has_attr_e.from DESC
    LIMIT 1

    // --------------
    // identify the active HAS_VALUE edges that we need to consider
    // --------------
    MATCH (attr)-[has_val_e:HAS_VALUE]->(av)
    WHERE (
        has_val_e.branch = $branch
        OR (has_val_e.branch_level < $branch_level AND has_val_e.from <= $branched_from)
    )
    AND has_val_e.status = "active"
    AND has_val_e.to IS NULL
    RETURN attr, has_val_e, av
    ORDER BY has_val_e.branch_level DESC, has_val_e.from DESC
    LIMIT 1
}

// --------------------
// determine the timestamp to index the AttributeValue
// --------------------
WITH attr, has_val_e, av,
CASE
    WHEN $from_time <= has_val_e.from THEN has_val_e.from
    ELSE $from_time
END AS indexed_from

// --------------------
// create the new edge to the AttributeValue vertex
// --------------------
WITH attr, has_val_e, av, indexed_from
CALL (attr, has_val_e, av, indexed_from) {
    WITH has_val_e
    WHERE NOT "AttributeValue" IN labels(av)

    MERGE (av_index:AttributeValue {value: av.value, is_default: av.is_default})
    LIMIT 1

    CREATE (attr)-[add_index_on_branch:HAS_VALUE]->(av_index)
    SET add_index_on_branch = properties(has_val_e)
    SET
        add_index_on_branch.branch = $branch,
        add_index_on_branch.branch_level = $branch_level,
        add_index_on_branch.from = indexed_from,
        add_index_on_branch.status = "active",
        add_index_on_branch.to = NULL

    // --------------------
    // delete existing active edge if it is on this branch and we created a new edge
    // --------------------
    WITH has_val_e
    WHERE has_val_e.branch = $branch AND add_index_on_branch IS NOT NULL
    DELETE has_val_e
}
        """ % {"schema_kind": self.schema_attribute_timeframe.kind}
        self.add_to_query(query)


class SetAttributeValueIndexedQuery(Query):
    name = "set_attribute_value_indexed_query"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (av:AttributeValue)
SET av:AttributeValueIndexed
        """
        self.add_to_query(query)


class FinalizeAttributeValueNonIndexedQuery(Query):
    name = "finalize_attribute_value_non_indexed_query"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
MATCH (av_no_index:AttributeValueNonIndexed)
SET av_no_index:AttributeValue
REMOVE av_no_index:AttributeValueNonIndexed
        """
        self.add_to_query(query)


class Migration037(ArbitraryMigration):
    """
    Update AttributeValue vertices to be AttributeValueIndexed, unless they include values for LARGE_ATTRIBUTE_TYPES

    0. Drop the index on the AttributeValueIndexed vertex, there are no AttributeValueIndexed vertices at this point anyway
    1. For all attributes of all schema on all branches, determine if the attribute is a LARGE_ATTRIBUTE_TYPE and when
        attribute's kind was last updated in the schema
    2. For all branches, starting with the default and global branches, update HAS_VALUE edges for LARGE_ATTRIBUTE_TYPE
        attributes to point to AttributeValueNonIndexed vertices
    3. For any LARGE_ATTRIBUTE_TYPE attributes on the default branch that were updated to non-large_type on other branches,
        revert the HAS_VALUE edges to point to AttributeValue vertices
    4. Add the AttributeValueIndexed label to all AttributeValue vertices
    5. Update all AttributeValueNonIndexed vertices to AttributeValue (no AttributeValueIndexed label)
    6. Any AttributeValueIndexed vertices with a value of size greater than MAX_STRING_LENGTH are changed to AttributeValueNonIndexed
    7. Add the index on AttributeValueIndexed again
    """

    name: str = "037_index_attr_vals"
    minimum_version: int = 36

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result

    async def execute(self, migration_input: MigrationInput) -> MigrationResult:  # noqa: PLR0915
        console = get_migration_console()
        db = migration_input.db
        result = MigrationResult()

        # find the active schema attributes that have a LARGE_ATTRIBUTE_TYPE kind on all branches
        console.print("Determining schema attribute types and timestamps on all branches...", end="")
        get_large_attribute_types_query = await GetLargeAttributeTypesQuery.init(db=db)
        await get_large_attribute_types_query.execute(db=db)
        schema_attribute_timeframes = get_large_attribute_types_query.get_large_attribute_type_timeframes()
        console.print("done")

        # find which schema attributes are large_types in the default branch, but updated to non-large_type on other branches
        # {(kind, attr_name): SchemaAttributeTimeframe}
        console.print("Determining schema attribute updates on non-default branches...", end="")
        main_schema_attribute_timeframes_map: dict[tuple[str, str], SchemaAttributeTimeframe] = {}
        for schema_attr_time in schema_attribute_timeframes:
            if schema_attr_time.is_default_branch:
                main_schema_attribute_timeframes_map[schema_attr_time.kind, schema_attr_time.attr_name] = (
                    schema_attr_time
                )
        large_type_reverts: list[SchemaAttributeTimeframe] = []
        for schema_attr_time in schema_attribute_timeframes:
            if schema_attr_time.is_default_branch or schema_attr_time.is_large_type:
                continue
            default_schema_attr_time = main_schema_attribute_timeframes_map.get(
                (schema_attr_time.kind, schema_attr_time.attr_name)
            )
            if not default_schema_attr_time:
                continue
            if (
                default_schema_attr_time.is_large_type
                and default_schema_attr_time.from_time < schema_attr_time.branched_from
            ):
                large_type_reverts.append(schema_attr_time)
        console.print("done")

        # drop the index on the AttributeValueNonIndexed vertex, there won't be any at this point anyway
        console.print("Dropping index on AttributeValueIndexed vertices...", end="")
        index_manager = IndexManagerNeo4j(db=db)
        index_manager.init(nodes=[AV_INDEXED_INDEX], rels=[])
        await index_manager.drop()
        console.print("done")

        # create the temporary non-indexed attribute value vertices for LARGE_ATTRIBUTE_TYPE attributes
        # start with default branch
        console.print("Creating temporary non-indexed attribute values for large attribute types...", end="")
        large_schema_attribute_timeframes = [
            schema_attr_time for schema_attr_time in schema_attribute_timeframes if schema_attr_time.is_large_type
        ]
        for schema_attr_time in sorted(large_schema_attribute_timeframes, key=lambda x: x.branch_level):
            create_non_indexed_attribute_value_query = await CreateNonIndexedAttributeValueQuery.init(
                db=db, schema_attribute_timeframe=schema_attr_time
            )
            await create_non_indexed_attribute_value_query.execute(db=db)
        console.print("done")

        # re-index attribute values on branches where the type was updated to non-large_type
        console.print("Re-indexing attribute values on branches updated to non-large types...", end="")
        for schema_attr_time in large_type_reverts:
            revert_non_index_on_branch_query = await RevertNonIndexOnBranchQuery.init(
                db=db, schema_attribute_timeframe=schema_attr_time
            )
            await revert_non_index_on_branch_query.execute(db=db)
        console.print("done")

        # set the AttributeValue vertices to be AttributeValueIndexed
        console.print("Adding AttributeValueIndexed label to AttributeValue vertices...", end="")
        set_attribute_value_indexed_query = await SetAttributeValueIndexedQuery.init(db=db)
        await set_attribute_value_indexed_query.execute(db=db)
        console.print("done")

        # set AttributeValueNonIndexed vertices to just AttributeValue
        console.print("Restoring AttributeValue label on AttributeValueNonIndexed vertices...", end="")
        finalize_attribute_value_non_indexed_query = await FinalizeAttributeValueNonIndexedQuery.init(db=db)
        await finalize_attribute_value_non_indexed_query.execute(db=db)
        console.print("done")

        # de-index all attribute values too large to be indexed
        console.print("De-indexing legacy attribute data exceeding index limits...", end="")
        de_index_large_attribute_values_query = await DeIndexLargeAttributeValuesQuery.init(
            db=db, max_value_size=MAX_STRING_LENGTH
        )
        await de_index_large_attribute_values_query.execute(db=db)
        console.print("done")

        # add the index back to the AttributeValueNonIndexed vertex
        console.print("Adding index back to the AttributeValueIndexed label...", end="")
        index_manager = IndexManagerNeo4j(db=db)
        index_manager.init(nodes=[AV_INDEXED_INDEX], rels=[])
        await index_manager.add()
        console.print("done")

        return result



from __future__ import annotations

from datetime import datetime
from typing import Any

from typing_extensions import NotRequired, TypedDict

from infrahub_sdk.graphql.query import Query


class NodeDiff(TypedDict):
    branch: str
    kind: str
    id: str
    action: str
    display_label: str
    elements: list[NodeDiffElement]


class NodeDiffElement(TypedDict):
    name: str
    element_type: str
    action: str
    summary: NodeDiffSummary
    peers: NotRequired[list[NodeDiffPeer]]


class NodeDiffSummary(TypedDict):
    added: int
    updated: int
    removed: int


class NodeDiffPeer(TypedDict):
    action: str
    summary: NodeDiffSummary


class DiffTreeData(TypedDict):
    num_added: int
    num_updated: int
    num_removed: int
    num_conflicts: int
    to_time: str
    from_time: str
    base_branch: str
    diff_branch: str
    name: NotRequired[str | None]
    nodes: list[NodeDiff]


def get_diff_summary_query() -> str:
    return """
        query GetDiffTree($branch_name: String!, $name: String, $from_time: DateTime, $to_time: DateTime) {
            DiffTree(branch: $branch_name, name: $name, from_time: $from_time, to_time: $to_time) {
                nodes {
                    uuid
                    kind
                    status
                    label
                    num_added
                    num_updated
                    num_removed
                    attributes {
                        name
                        status
                        num_added
                        num_updated
                        num_removed
                    }
                    relationships {
                        name
                        status
                        cardinality
                        num_added
                        num_updated
                        num_removed
                        elements {
                            status
                            num_added
                            num_updated
                            num_removed
                        }
                    }
                }
            }
        }
    """


def diff_tree_node_to_node_diff(node_dict: dict[str, Any], branch_name: str) -> NodeDiff:
    element_diffs: list[NodeDiffElement] = []
    if "attributes" in node_dict:
        for attr_dict in node_dict["attributes"]:
            attr_diff = NodeDiffElement(
                action=str(attr_dict.get("status")),
                element_type="ATTRIBUTE",
                name=str(attr_dict.get("name")),
                summary={
                    "added": int(attr_dict.get("num_added") or 0),
                    "removed": int(attr_dict.get("num_removed") or 0),
                    "updated": int(attr_dict.get("num_updated") or 0),
                },
            )
            element_diffs.append(attr_diff)
    if "relationships" in node_dict:
        for relationship_dict in node_dict["relationships"]:
            is_cardinality_one = str(relationship_dict.get("cardinality")).upper() == "ONE"
            relationship_diff = NodeDiffElement(
                action=str(relationship_dict.get("status")),
                element_type="RELATIONSHIP_ONE" if is_cardinality_one else "RELATIONSHIP_MANY",
                name=str(relationship_dict.get("name")),
                summary={
                    "added": int(relationship_dict.get("num_added") or 0),
                    "removed": int(relationship_dict.get("num_removed") or 0),
                    "updated": int(relationship_dict.get("num_updated") or 0),
                },
            )
            if not is_cardinality_one and "elements" in relationship_dict:
                peer_diffs = [
                    NodeDiffPeer(
                        action=str(element_dict.get("status")),
                        summary={
                            "added": int(element_dict.get("num_added") or 0),
                            "removed": int(element_dict.get("num_removed") or 0),
                            "updated": int(element_dict.get("num_updated") or 0),
                        },
                    )
                    for element_dict in relationship_dict["elements"]
                ]
                relationship_diff["peers"] = peer_diffs
            element_diffs.append(relationship_diff)
    return NodeDiff(
        branch=branch_name,
        kind=str(node_dict.get("kind")),
        id=str(node_dict.get("uuid")),
        action=str(node_dict.get("status")),
        display_label=str(node_dict.get("label")),
        elements=element_diffs,
    )


def get_diff_tree_query() -> Query:
    node_structure = {
        "uuid": None,
        "kind": None,
        "status": None,
        "label": None,
        "num_added": None,
        "num_updated": None,
        "num_removed": None,
        "attributes": {
            "name": None,
            "status": None,
            "num_added": None,
            "num_updated": None,
            "num_removed": None,
        },
        "relationships": {
            "name": None,
            "status": None,
            "cardinality": None,
            "num_added": None,
            "num_updated": None,
            "num_removed": None,
            "elements": {
                "status": None,
                "num_added": None,
                "num_updated": None,
                "num_removed": None,
            },
        },
    }

    return Query(
        name="GetDiffTree",
        query={
            "DiffTree": {
                "@filters": {
                    "branch": "$branch_name",
                    "name": "$name",
                    "from_time": "$from_time",
                    "to_time": "$to_time",
                },
                "name": None,
                "to_time": None,
                "from_time": None,
                "base_branch": None,
                "diff_branch": None,
                "num_added": None,
                "num_updated": None,
                "num_removed": None,
                "num_conflicts": None,
                "nodes": node_structure,
            },
        },
        variables={
            "branch_name": str,
            "name": str | None,
            "from_time": datetime | None,
            "to_time": datetime | None,
        },
    )



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from prefect.events import emit_event

if TYPE_CHECKING:
    from infrahub.events import InfrahubEvent
    from infrahub.services.adapters.message_bus import InfrahubMessageBus


class InfrahubEventService:
    """Base class for infrahub event service"""

    def __init__(self, message_bus: InfrahubMessageBus | None = None) -> None:
        # Ideally message_bus should not be optional, we let it like this for existing tests that
        #  pass without a bus as corresponding tested events do not send bus messages.
        self.message_bus = message_bus

    async def send(self, event: InfrahubEvent) -> None:
        tasks = [self._send_bus(event=event), self._send_prefect(event=event)]
        await asyncio.gather(*tasks)

    async def _send_bus(self, event: InfrahubEvent) -> None:
        for message in event.get_messages():
            if self.message_bus is None:
                raise ValueError("InfrahubEventService.message_bus is None.")
            await self.message_bus.send(message=message)

    async def _send_prefect(self, event: InfrahubEvent) -> None:
        emit_event(
            id=event.meta.id,
            event=event.event_name,
            resource=event.get_resource(),
            related=event.get_related(),
            payload=event.get_event_payload(),
        )



from abc import ABC, abstractmethod

from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes

from ..model import RelationshipManager


class RelationshipManagerConstraintInterface(ABC):
    @abstractmethod
    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes, node: Node) -> None: ...

    def expand_filters(self, field_filters: list[str], node_schema: MainSchemaTypes) -> list[str]:  # noqa: ARG002
        """Expand field_filters to include additional relationships this constraint needs checked. Override in
        subclasses that need cross-field validation."""
        return field_filters



from prefect import flow

from infrahub.artifacts.models import CheckArtifactCreate
from infrahub.core.constants import InfrahubKind, ValidatorConclusion
from infrahub.core.timestamp import Timestamp
from infrahub.git.repository import get_initialized_repo
from infrahub.tasks.artifact import define_artifact
from infrahub.workers.dependencies import get_client
from infrahub.workflows.utils import add_tags


@flow(name="git-repository-check-artifact-create", flow_run_name="Check artifact creation")
async def create(model: CheckArtifactCreate) -> ValidatorConclusion:
    await add_tags(branches=[model.branch_name], nodes=[model.target_id])

    client = get_client()
    client.request_context = model.context.to_request_context()

    validator = await client.get(kind=InfrahubKind.ARTIFACTVALIDATOR, id=model.validator_id, include=["checks"])

    repo = await get_initialized_repo(
        client=client,
        repository_id=model.repository_id,
        name=model.repository_name,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )

    artifact, artifact_created = await define_artifact(model=model)

    severity = "info"
    artifact_result: dict[str, str | bool | None] = {
        "changed": None,
        "checksum": None,
        "artifact_id": None,
        "storage_id": None,
    }
    check_message = "Failed to render artifact"

    try:
        result = await repo.render_artifact(artifact=artifact, artifact_created=artifact_created, message=model)
        artifact_result["changed"] = result.changed
        artifact_result["checksum"] = result.checksum
        artifact_result["artifact_id"] = result.artifact_id
        artifact_result["storage_id"] = result.storage_id
        check_message = "Artifact rendered successfully"
        conclusion = ValidatorConclusion.SUCCESS

    except Exception as exc:
        artifact.status.value = "Error"
        await artifact.save()
        severity = "critical"
        conclusion = ValidatorConclusion.FAILURE
        check_message += f": {str(exc)}"

    check = None
    check_name = f"{model.artifact_name}: {model.target_name}"
    existing_check = await client.filters(
        kind=InfrahubKind.ARTIFACTCHECK, validator__ids=validator.id, name__value=check_name
    )
    if existing_check:
        check = existing_check[0]

    if check:
        check.created_at.value = Timestamp().to_string()
        check.conclusion.value = conclusion.value
        check.severity.value = severity
        check.changed.value = artifact_result["changed"]
        check.checksum.value = artifact_result["checksum"]
        check.artifact_id.value = artifact_result["artifact_id"]
        check.storage_id.value = artifact_result["storage_id"]
        await check.save()
    else:
        check = await client.create(
            kind=InfrahubKind.ARTIFACTCHECK,
            data={
                "name": check_name,
                "origin": model.repository_id,
                "kind": "ArtifactDefinition",
                "validator": model.validator_id,
                "created_at": Timestamp().to_string(),
                "message": check_message,
                "conclusion": conclusion.value,
                "severity": severity,
                "changed": artifact_result["changed"],
                "checksum": artifact_result["checksum"],
                "artifact_id": artifact_result["artifact_id"],
                "storage_id": artifact_result["storage_id"],
            },
        )
        await check.save()

    return conclusion



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub.core.creation_context import NodeCreationContext
from infrahub.pools.allocator import PoolAllocator

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase


class DefaultPoolAllocator(PoolAllocator):
    """Pool allocator for regular node creation. Allocates from number pools and IP/Prefix pools."""

    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch

    async def allocate_for_attribute(
        self,
        pool_relationship: RelationshipManager,
        target_schema: MainSchemaTypes,
        attribute_name: str,
        identifier: str,
    ) -> Any | None:
        """Allocate from a _from_resource_pool relationship for an attribute (Number → CoreNumberPool)."""
        pool = await pool_relationship.get_peer(db=self.db)
        if not pool:
            return None

        attr_schema = target_schema.get_attribute(name=attribute_name)
        return await pool.get_resource(db=self.db, branch=self.branch, identifier=identifier, attribute=attr_schema)  # type: ignore[attr-defined]

    async def allocate_for_relationship(self, pool_relationship: RelationshipManager, identifier: str) -> Node | None:
        """Allocate from pool referenced by relationship."""
        pool = await pool_relationship.get_peer(db=self.db)
        if not pool:
            return None

        allocated = await pool.get_resource(db=self.db, branch=self.branch, identifier=identifier)  # type: ignore
        if allocated:
            NodeCreationContext.record_if_active(node=allocated)
        return allocated



from .artifact_action import ArtifactCreatedEvent, ArtifactUpdatedEvent
from .branch_action import BranchCreatedEvent, BranchDeletedEvent, BranchMergedEvent, BranchRebasedEvent
from .group_action import GroupMemberAddedEvent, GroupMemberRemovedEvent
from .models import EventMeta, InfrahubEvent
from .node_action import NodeCreatedEvent, NodeDeletedEvent, NodeUpdatedEvent
from .proposed_change_action import (
    ProposedChangeApprovalRevokedEvent,
    ProposedChangeApprovalsRevokedEvent,
    ProposedChangeApprovedEvent,
    ProposedChangeMergedEvent,
    ProposedChangeRejectedEvent,
    ProposedChangeRejectionRevokedEvent,
    ProposedChangeReviewRequestedEvent,
    ProposedChangeThreadCreatedEvent,
    ProposedChangeThreadUpdatedEvent,
)
from .repository_action import CommitUpdatedEvent
from .validator_action import ValidatorFailedEvent, ValidatorPassedEvent, ValidatorStartedEvent

__all__ = [
    "ArtifactCreatedEvent",
    "ArtifactUpdatedEvent",
    "BranchCreatedEvent",
    "BranchDeletedEvent",
    "BranchMergedEvent",
    "BranchRebasedEvent",
    "CommitUpdatedEvent",
    "EventMeta",
    "GroupMemberAddedEvent",
    "GroupMemberRemovedEvent",
    "InfrahubEvent",
    "NodeCreatedEvent",
    "NodeDeletedEvent",
    "NodeUpdatedEvent",
    "ProposedChangeApprovalRevokedEvent",
    "ProposedChangeApprovalsRevokedEvent",
    "ProposedChangeApprovedEvent",
    "ProposedChangeMergedEvent",
    "ProposedChangeRejectedEvent",
    "ProposedChangeRejectionRevokedEvent",
    "ProposedChangeReviewRequestedEvent",
    "ProposedChangeThreadCreatedEvent",
    "ProposedChangeThreadUpdatedEvent",
    "ValidatorFailedEvent",
    "ValidatorPassedEvent",
    "ValidatorStartedEvent",
]



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar

from infrahub.core.branch import Branch
from infrahub.database import InfrahubDatabase

T = TypeVar("T")


@dataclass
class DependencyBuilderContext:
    db: InfrahubDatabase
    branch: Branch


class DependencyBuilder[T](ABC):
    @classmethod
    @abstractmethod
    def build(cls, context: DependencyBuilderContext) -> T: ...
